/*
 * LCD peripheral support for the JZ4740 and related SoCs.
 *
 * Copyright (C) 2018, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include "lcd.h"
#include "lcd-jz4740-panel.h"

#include <l4/sys/types.h>

#include <stdint.h>

/* Descriptor referenced by the DMA mechanism extended to 8 words for JZ4780
   support. */

struct Jz4740_lcd_descriptor
{
  struct Jz4740_lcd_descriptor *next; /* FDADR: frame descriptor address */
  uint32_t source;                    /* FSADR: frame source address */
  uint32_t identifier;                /* FIDR: frame identifier */
  uint32_t command;                   /* CMD: command */

  /* "New" descriptor fields. */

  uint32_t offset;                    /* OFFSIZE: offset in words between lines */
  uint32_t page_width;                /* PW: number of words per line (16x16 block mode) */
  uint32_t command_position;          /* CNUM: command number (smart LCD mode) or
                                         CPOS: foreground position and properties */
  uint32_t fg_size;                   /* DESSIZE: foreground size and alpha properties */
};

/* Interrupt conditions. */

enum Jz4740_lcd_irq_condition
{
  Lcd_irq_none        = 0,
  Lcd_irq_frame_start = 1,
  Lcd_irq_frame_end   = 2,
};



/* C++ language interface. */

#ifdef __cplusplus

#include <l4/devices/hw_mmio_register_block.h>

/* General JZ4740 LCD controller support. */

class Lcd_jz4740_chip : public Lcd_chip
{
protected:
  Hw::Register_block<32> _regs;
  Jz4740_lcd_panel *_panel;
  int _burst_size;
  enum Jz4740_lcd_irq_condition _irq_conditions = Lcd_irq_none;
  l4_cap_idx_t _irq = L4_INVALID_CAP;

  /* Control register value calculation. */

  uint32_t _control_bpp();
  uint32_t _control_irq();
  uint32_t _control_panel();
  uint32_t _control_stn_frc();
  uint32_t _control_transfer();

  /* OSD configure register value calculation. */

  uint32_t _osd_config_irq();

  /* Command register value calculation. */

  uint32_t _command_irq();

  /* Status register value calculation. */

  uint32_t _status_irq();

  /* Panel mode access. */

  uint32_t _mode();

  /* Panel initialisation. */

  void _init_stn();
  void _init_tft();
  void _init_panel();

  /* Descriptor initialisation. */

  virtual void _set_descriptor(struct Jz4740_lcd_descriptor &desc, l4_addr_t source,
                               l4_size_t size, struct Jz4740_lcd_descriptor *next,
                               uint32_t flags = 0);

public:
  Lcd_jz4740_chip(l4_addr_t addr, Jz4740_lcd_panel *panel);

  struct Jz4740_lcd_panel *get_panel();

  /* Peripheral control. */

  void disable();
  void disable_quick();
  void enable();
  bool enabled();

  /* Peripheral properties. */

  int get_pixel_clock();

  /* Panel properties. */

  int get_panels();
  int have_stn_panel();
  int have_colour_stn();
  int have_serial_tft();

  /* Memory properties. */

  l4_size_t get_pixel_size();
  l4_size_t get_line_size();
  l4_size_t get_screen_size();
  l4_size_t get_aligned_size();
  l4_size_t get_palette_size();
  l4_size_t get_aligned_palette_size();

  /* Memory properties for allocation purposes. */

  l4_size_t get_total_size();
  l4_size_t get_descriptors_size();

  /* Memory region access. */

  l4_addr_t get_palette(l4_addr_t screen);
  l4_addr_t get_framebuffer(int panel, l4_addr_t screen);

  /* Convenience methods. */

  void init_palette(l4_addr_t palette);

  /* Configuration. */

  virtual void config(struct Jz4740_lcd_descriptor *desc_vaddr,
                      struct Jz4740_lcd_descriptor *desc_paddr,
                      l4_addr_t fb_paddr);

  /* Interrupt configuration. */

  void set_irq(l4_cap_idx_t irq, enum Jz4740_lcd_irq_condition conditions);

  /* Interrupt handling. */

  long wait_for_irq();
};

#endif



/* C language interface. */

EXTERN_C_BEGIN

void *jz4740_lcd_init(l4_addr_t lcd_base, struct Jz4740_lcd_panel *panel);

void jz4740_lcd_config(void *lcd, struct Jz4740_lcd_descriptor *desc_vaddr,
                                  struct Jz4740_lcd_descriptor *desc_paddr,
                                  l4_addr_t fb_paddr);

void jz4740_lcd_set_irq(void *lcd, l4_cap_idx_t irq,
                        enum Jz4740_lcd_irq_condition conditions);

long jz4740_lcd_wait_for_irq(void *lcd);

void jz4740_lcd_disable(void *lcd);
void jz4740_lcd_disable_quick(void *lcd);
void jz4740_lcd_enable(void *lcd);
int jz4740_lcd_enabled(void *lcd);

l4_size_t jz4740_lcd_get_descriptors_size(void *lcd);
l4_size_t jz4740_lcd_get_line_size(void *lcd);
l4_size_t jz4740_lcd_get_screen_size(void *lcd);

l4_addr_t jz4740_lcd_get_palette(void *lcd, l4_addr_t base);
int jz4740_lcd_get_pixel_clock(void *lcd);

void jz4740_lcd_init_palette(void *lcd, l4_addr_t palette);

EXTERN_C_END
